#include <iostream>
#include <vector>
#include <cmath>
#include "omp.h"
using namespace std ;

//using integer=int ; // 32 bits integer
using integer=long ; // 64 bits integer

double dot_product(const integer n,const double *x,const double *y)
{
  double dp =0.0 ;
  //#pragma omp parallel for reduction(+:ds)
  #pragma acc parallel loop present(x[0:n],y[0:n]) reduction(+:dp)
  for(integer i=0;i<n;++i)
    dp += x[i]*y[i] ;
  return dp ;
}

int main(int argc,char**argv)
{
  integer N = 10 ;
  integer M = 1 ;
  integer outVec = 0;
  
  double ta,tb ;
  double tdp = 0.0;

  if(argc==2 && ( argv[1]==string("-h") ||  argv[1]==string("--help")) )
  {
    cout << "Usage: " << argv[0] << " [ -n integer, -m integer, -o integer ]" << endl;
    return 0;
  }

  for(int i=0;i<argc;++i)
    cout << argv[i] << " ";
  cout << endl;


  for(int i=0;i<argc;++i)
  {
    if(argv[i]==string("-n"))
      N = stol(argv[i+1]);
    if(argv[i]==string("-m"))
      M = stol(argv[i+1]);
    if(argv[i]==string("-o"))
      outVec = stol(argv[i+1]);
  }

  cout << "vectors size (-n " << N << ") dot product iterations (-m " << M << ")" << endl;
  
  ta = omp_get_wtime();
  vector<double> x(N) ;
  vector<double> y(N) ;
  double * px = &x[0];
  double * py = &y[0];
  tb = omp_get_wtime();
  cout << "vectors allocated in " << tb-ta << " seconds" << endl;
  //#pragma acc enter data create(x[0:N],y[0:N]) copyin(tdp)
  //#pragma acc data copyin(tdp,px[0:N],py[0:N]) copyout(tdp)
  //#pragma acc data copyin(tdp,px[0:N],py[0:N]) copyout(tdp,px[0:N],py[0:N])
  #pragma acc data create(px[0:N],py[0:N]) copy(tdp)
  { // BEGIN ACC DATA REGION
  
  ta = omp_get_wtime();
  //#pragma acc parallel loop copyin(px[0:N],py[0:N])
  #pragma acc parallel loop present(px[0:N],py[0:N])
  for(integer i=0;i<N;++i)
  {
    px[i] = sin(i) ;
    py[i] = cos(i) ;
  }
  tb = omp_get_wtime();
  cout << "vectors initialized in " << tb-ta << " seconds" << endl;

  ta = omp_get_wtime();
  for(int i=0;i<M;++i)
  {
    // tdp += dot_product(N,px,py);
    #pragma acc parallel loop present(px[0:N],py[0:N]) reduction(+:tdp)
    for(integer i=0;i<N;++i)
      tdp += px[i]*py[i] ;
  }
  tb = omp_get_wtime();
 
  } // END ACC DATA REGION

  if(outVec)
  {
    int limit = N;
    if(outVec>0)
        limit = outVec;
  
    for(int i=0;i<limit;++i)
      cout << x[i] << " ";
    cout << endl;
    for(int i=0;i<limit;++i)
    cout << y[i] << " ";
  cout << endl;
  }
  cout << "m dot products (" << tdp << ") computed in " << tb-ta << " seconds" << endl;
  
  return 0 ;
}
